#include <stdlib.h>
#include <stdio.h>

#include "list.h"

list_t *list(void *hd, list_t *tl) 
{ list_t *l;
  if ((l= (list_t*)malloc(sizeof(list_t)))== NULL) {
    printf("Out of memory\n"); exit(-1);
  }
  l-> elem= hd; l-> next= tl;
  return l;
}

int length(list_t * l)
{
  int i;
 
  for (i= 0; l!= NULL; l= l-> next) i++;

  return i;
}


/* Filter, rekursiv -- erzeugt neue Liste */
list_t *filter(int f(void *x), list_t *l)
{ 
  list_t *r;

  if (l != NULL) {
    r= filter(f, l-> next);
    if (f(l-> elem)) {
      return list(l-> elem, r);
    }       
    else {
      return r;
    } 
  }
  return NULL;
}


/* Map, rekursiv -- erzeugt neue Liste */
list_t *map(void *f(void *x), list_t *l)
{
  list_t *n;

  if (l != NULL) {
    n= map(f, l-> next);
    return list(f(l-> elem), n);
  } 
  return NULL;
}  

/* Map, iterative -- ersetzt Liste in-place */
list_t *map1(void *f(void *x), list_t *l)
{
  list_t *c;
  for (c= l; c!= NULL; c= c-> next) {
    c-> elem= f(c-> elem); }
  return l;
}   

/* NB: Hat ein Speicherleck -- die alten c-> elem ist nicht mehr erreichbar.
 * Wenn man das ändert, wird es länglich und es ist nicht ganz ungefährlich --
 *  c-> elem könte auch von anderswo referenziert werden, dann ist das free gefährlich.
 */
list_t *map2(void *f(void *x), list_t *l)
{
  list_t *c; void *d;  
  for (c= l; c!= NULL; c= c-> next) {
    d= c-> elem;
    c-> elem= f(c-> elem);
    free(d);
    }
  return l;
}   


/* foreach oder mapM_ -- map nur mit Seiteneffekten */
void foreach(void f(void *x), list_t *l)
{
  list_t *c;
  for (c= l; c!= NULL; c= c-> next) {
    f(c-> elem);
    }
}

/* Generische Ausgabefunktion */
void print_list(list_t *l, void print_elem(void *))
{
  int tmp= 0;
  
  printf("[");
  while (l != NULL) {
    if (tmp) printf(", "); else tmp= 1;
    print_elem(l->elem);
    l= l-> next;
  }
  printf("]\n");
}
